use std::{fmt::Display, time::Duration};

use ev3dev_lang_rust::Screen;
use image::Rgb;
use imageproc::{
    drawing::{draw_line_segment_mut, draw_polygon_mut, draw_text_mut, text_size},
    point::Point,
};
use rusttype::{Font, Scale};

const BACKGROUND: Rgb<u8> = Rgb([255; 3]);
const FOREGROUND: Rgb<u8> = Rgb([0; 3]);
const FONT_DATA: &[u8] = include_bytes!("../fonts/RobotoMono-Regular.ttf");

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone, Copy)]
pub enum Parameter {
    Kp,
    Ki,
    Kd,
    Speed,
}

pub fn draw_settings(
    screen: &mut Screen,
    kp: impl Display,
    ki: impl Display,
    kd: impl Display,
    v: impl Display,
    selected: Parameter,
    font: &Font,
) {
    let text_scale = Scale::uniform(screen.yres() as f32 / 4.0 - 8.0);
    clear(screen);
    draw_centered_text(
        screen,
        screen.xres() as i32 / 2,
        4,
        text_scale,
        font,
        "Settings",
    );

    draw_setting(screen, kp, selected == Parameter::Kp, 0, "k_p", font);
    draw_setting(screen, ki, selected == Parameter::Ki, 1, "k_i", font);
    draw_setting(screen, kd, selected == Parameter::Kd, 2, "k_d", font);
    draw_setting(screen, v, selected == Parameter::Speed, 3, "v", font);

    screen.update();
}

pub fn draw_driving(screen: &mut Screen, font: &Font) {
    clear(screen);
    draw_centered_text(
        screen,
        screen.xres() as i32 / 2,
        4,
        Scale::uniform(screen.yres() as f32 / 2.0 - 8.0),
        font,
        "Fahrt",
    );
    screen.update();
}

pub fn draw_finished(screen: &mut Screen, time: Duration, font: &Font) {
    clear(screen);
    let text_scale = Scale::uniform(screen.yres() as f32 / 2.0 - 8.0);
    draw_centered_text(
        screen,
        screen.xres() as i32 / 2,
        4,
        text_scale,
        font,
        "Zeit:",
    );
    draw_centered_text(
        screen,
        screen.xres() as i32 / 2,
        screen.yres() as i32 / 2 + 4,
        text_scale,
        font,
        format!("{:.2}s", time.as_secs_f32()).as_str(),
    );
    screen.update();
}

pub fn draw_cycles(screen: &mut Screen, cycles: u32, font: &Font) {
    clear(screen);
    let text_scale = Scale::uniform(screen.yres() as f32 / 2.0 - 8.0);
    draw_centered_text(
        screen,
        screen.xres() as i32 / 2,
        4,
        text_scale,
        font,
        "Cycles:",
    );
    draw_centered_text(
        screen,
        screen.xres() as i32 / 2,
        screen.yres() as i32 / 2 + 4,
        text_scale,
        font,
        format!("{cycles}").as_str(),
    );
    screen.update();
}

pub fn font() -> Font<'static> {
    Font::try_from_bytes(FONT_DATA).unwrap()
}

fn draw_setting(
    screen: &mut Screen,
    value: impl Display,
    selected: bool,
    index: u8,
    name: &str,
    font: &Font,
) {
    let text_scale = Scale::uniform(screen.yres() as f32 / 4.0 - 8.0);
    let x = screen.xres() as i32 / 8 + screen.xres() as i32 * i32::from(index) / 4;
    draw_centered_text(
        screen,
        x,
        screen.yres() as i32 / 4 + 4,
        text_scale,
        font,
        name,
    );
    draw_centered_text(
        screen,
        x,
        screen.yres() as i32 * 5 / 8 + 4,
        text_scale,
        font,
        format!("{value}").as_str(),
    );
    let top_triangle = [
        Point::new(x, screen.yres() as i32 / 2),
        Point::new(x - 10, screen.yres() as i32 / 2 + 10),
        Point::new(x + 10, screen.yres() as i32 / 2 + 10),
    ];
    let bottom_triangle = [
        Point::new(x, screen.yres() as i32),
        Point::new(x - 10, screen.yres() as i32 - 10),
        Point::new(x + 10, screen.yres() as i32 - 10),
    ];
    if selected {
        draw_polygon_mut(&mut screen.image, &top_triangle, FOREGROUND);
        draw_polygon_mut(&mut screen.image, &bottom_triangle, FOREGROUND);
    } else {
        for (start, end) in [
            (top_triangle[0], top_triangle[1]),
            (top_triangle[1], top_triangle[2]),
            (top_triangle[2], top_triangle[0]),
            (bottom_triangle[0], bottom_triangle[1]),
            (bottom_triangle[1], bottom_triangle[2]),
            (bottom_triangle[2], bottom_triangle[0]),
        ] {
            draw_line_segment_mut(
                &mut screen.image,
                point2tuple(start),
                point2tuple(end),
                FOREGROUND,
            );
        }
    }
}

fn draw_centered_text(
    screen: &mut ev3dev_lang_rust::Screen,
    x: i32,
    y: i32,
    scale: Scale,
    font: &Font,
    text: &str,
) {
    let (width, _) = text_size(scale, font, text);
    draw_text_mut(
        &mut screen.image,
        FOREGROUND,
        x - width / 2,
        y,
        scale,
        font,
        text,
    );
}

fn clear(screen: &mut Screen) {
    screen.image.fill(BACKGROUND.0[0]);
}

const fn point2tuple(point: Point<i32>) -> (f32, f32) {
    #[allow(clippy::cast_precision_loss)]
    (point.x as f32, point.y as f32)
}
